# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from enum import Enum
from argcomplete.completers import FilesCompleter

from knack.arguments import CLIArgumentType

from azure.mgmt.batch.models import (
    AccountKeyType,
    KeySource,
    PublicNetworkAccessType,
    ResourceIdentityType,
    EndpointAccessDefaultAction)
from azure.batch.models import (
    CachingType,
    BatchNodeDeallocationOption,
    StorageAccountType)

from azure.cli.core.commands.parameters import (
    file_type,
    get_enum_type,
    get_location_type,
    get_resource_name_completion_list,
    get_three_state_flag,
    resource_group_name_type,
    tags_type)

from azure.cli.command_modules.batch._completers import load_supported_images
from azure.cli.command_modules.batch._validators import (
    application_enabled,
    batch_application_package_reference_format,
    datetime_format,
    disk_encryption_configuration_format,
    environment_setting_format,
    keyvault_id,
    metadata_item_format,
    resource_file_format,
    duration_format,
    storage_account_id,
    validate_client_parameters,
    validate_json_file,
    validate_pool_resize_parameters)


class NetworkProfile(Enum):
    BatchAccount = "BatchAccount"
    NodeManagement = "NodeManagement"


profile_param_type = CLIArgumentType(
    help="Network profile to set.",
    arg_type=get_enum_type(NetworkProfile))


# pylint: disable=line-too-long, too-many-statements
def load_arguments(self, _):
    batch_name_type = CLIArgumentType(
        help='Name of the Batch account.',
        options_list=('--account-name',),
        completer=get_resource_name_completion_list('Microsoft.Batch/batchAccounts'),
        id_part=None)

    with self.argument_context('batch') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group', completer=None, validator=None, required=True)

    with self.argument_context('batch account') as c:
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'))

    with self.argument_context('batch account show') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group. If not specified will display currently set account.', required=False)
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'), help='Name of the batch account to show. If not specified will display currently set account.', required=False)

    with self.argument_context('batch account list') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group', required=False)

    with self.argument_context('batch account create') as c:
        c.argument('location', get_location_type(self.cli_ctx), help='The region in which to create the account.')
        c.argument('tags', tags_type, help="Space-separated tags in 'key[=value]' format.")
        c.argument('storage_account', help='The storage account name or resource ID to be used for auto storage.', validator=storage_account_id)
        c.argument('keyvault', help='The KeyVault name or resource ID to be used for an account with a pool allocation mode of \'User Subscription\'.', validator=keyvault_id)
        c.argument('public_network_access', help="The network access type for accessing Azure Batch account. Values can either be enabled or disabled.", arg_type=get_enum_type(PublicNetworkAccessType))
        c.argument('encryption_key_source', help='Part of the encryption configuration for the Batch account. Type of the key source. Can be either Microsoft.Batch or Microsoft.KeyVault', arg_type=get_enum_type(KeySource))
        c.argument('encryption_key_identifier', help='Part of the encryption configuration for the Batch account. '
                                                     'Full path to the versioned secret. Example https://mykeyvault.vault.azure.net/keys/testkey/6e34a81fef704045975661e297a4c053.')
        c.argument('identity_type', help="The type of identity used for the Batch account. Possible values include: 'SystemAssigned', 'None'.", arg_type=get_enum_type(ResourceIdentityType), deprecate_info=c.deprecate(hide=True))
        c.argument('mi_user_assigned', help='Resource ID of the user assigned identity for the batch services account.', arg_group='Identity')
        c.argument('mi_system_assigned', help='Set the system managed identity on the batch services account.', arg_group='Identity')
        c.ignore('keyvault_url')

    with self.argument_context('batch account set') as c:
        c.argument('tags', tags_type)
        c.argument('storage_account', help='The storage account name or resource ID to be used for auto storage.', validator=storage_account_id)
        c.argument('encryption_key_source', help='Part of the encryption configuration for the Batch account. Type of the key source. Can be either Microsoft.Batch or Microsoft.KeyVault')
        c.argument('public_network_access', help="The network access type for accessing Azure Batch account. Values can either be enabled or disabled.", arg_type=get_enum_type(PublicNetworkAccessType))
        c.argument('encryption_key_identifier', help='Part of the encryption configuration for the Batch account. Full path to the versioned secret. Example https://mykeyvault.vault.azure.net/keys/testkey/6e34a81fef704045975661e297a4c053.')
        c.argument('identity_type', help="The type of identity used for the Batch account. Possible values include: 'SystemAssigned', 'None'.", arg_type=get_enum_type(ResourceIdentityType), deprecate_info=c.deprecate(hide=True))

    with self.argument_context('batch account identity assign') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group. If not specified will display currently set account.')
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'), help='Name of the batch account to show. If not specified will display currently set account.')
        c.argument('mi_system_assigned', options_list=['--system-assigned'], action='store_true',
                   arg_group='Managed Identity', help='Provide this flag to use system assigned identity for batch accounts. '
                   'Check out help for more examples')
        c.argument('mi_user_assigned', options_list=['--user-assigned'],
                   arg_group='Managed Identity', help='User Assigned Identity ids to be used for batch account. '
                   'Check out help for more examples')

    with self.argument_context('batch account identity remove') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group. If not specified will display currently set account.')
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'), help='Name of the batch account to show. If not specified will display currently set account.')
        c.argument('mi_system_assigned', options_list=['--system-assigned'], action='store_true',
                   arg_group='Managed Identity', help='Provide this flag to use system assigned identity for batch accounts. '
                   'Check out help for more examples')
        c.argument('mi_user_assigned', options_list=['--user-assigned'], nargs='*',
                   arg_group='Managed Identity', help='User Assigned Identity ids to be used for batch account. '
                   'Check out help for more examples')

    with self.argument_context('batch account identity show') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group. If not specified will display currently set account.')
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'), help='Name of the batch account to show. If not specified will display currently set account.')

    with self.argument_context('batch account network-profile show') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group. If not specified will display currently set account.')
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'), help='Name of the batch account to show. If not specified will display currently set account.')

    with self.argument_context('batch account network-profile set') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group. If not specified will display currently set account.')
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'), help='Name of the batch account to show. If not specified will display currently set account.')
        c.argument('profile', arg_type=profile_param_type)
        c.argument('default_action', help="Default action for endpoint access. It is only applicable when publicNetworkAccess is enabled. Possible values include: 'Allow', 'Deny'", arg_type=get_enum_type(EndpointAccessDefaultAction))

    with self.argument_context('batch account network-profile network-rule list') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group. If not specified will display currently set account.')
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'), help='Name of the batch account to show. If not specified will display currently set account.')

    with self.argument_context('batch account network-profile network-rule add') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group. If not specified will display currently set account.')
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'), help='Name of the batch account to show. If not specified will display currently set account.')
        c.argument('profile', arg_type=profile_param_type)
        c.argument('ip_address', help='IPv4 address or CIDR range.')

    with self.argument_context('batch account network-profile network-rule delete') as c:
        c.argument('resource_group_name', resource_group_name_type, help='Name of the resource group. If not specified will display currently set account.')
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'), help='Name of the batch account to show. If not specified will display currently set account.')
        c.argument('profile', arg_type=profile_param_type)
        c.argument('ip_address', help='IPv4 address or CIDR range.')

    with self.argument_context('batch account keys renew') as c:
        c.argument('resource_group_name', resource_group_name_type,
                   help='Name of the resource group. If not specified will display currently set account.',
                   required=False)
        c.argument('account_name', batch_name_type, options_list=('--name', '-n'),
                   help='Name of the batch account to show. If not specified will display currently set account.',
                   required=False)
        c.argument('key_name', arg_type=get_enum_type(AccountKeyType), help='Name of the batch account key.')

    with self.argument_context('batch account login') as c:
        c.argument('shared_key_auth', action='store_true', help='Using Shared Key authentication, if not specified, it will use Azure Active Directory authentication.')
        c.argument('show', action='store_true', help='Display the credential information for the Batch account.')

    with self.argument_context('batch application set') as c:
        c.argument('application_name', help="The name of the application.")
        c.argument('allow_updates', help="Specify to indicate whether packages within the application may be overwritten using the same version string. Specify either 'true' or 'false' to update the property.")
        c.argument('default_version', help="Specify which package to use if a client requests the application but does not specify a version.")
        c.argument('display_name', help="Specify the display name for the application.")

    with self.argument_context('batch application create') as c:
        c.argument('allow_updates', options_list=('--allow-updates',), action="store_true", help="Specify to indicate whether packages within the application may be overwritten using the same version string. True if flag present.")

    for command in ['create', 'activate']:
        with self.argument_context(f'batch application package {command}') as c:
            c.argument('package_file', type=file_type, help='The path of the application package in zip format', completer=FilesCompleter())
            c.argument('application_name', help="The name of the application.")
            c.argument('version_name', help="The version name of the application.")
            c.argument('f_ormat', options_list=('--format',), help="The format of the application package binary file.")

    with self.argument_context('batch location quotas show') as c:
        c.argument('location_name', get_location_type(self.cli_ctx), help='The region for which to display the Batch service quotas.')

    with self.argument_context('batch location list-skus') as c:
        c.argument('location_name', get_location_type(self.cli_ctx), help='The region for which to display the available Batch VM SKUs.')

    for command in ['list', 'show', 'create', 'set', 'delete', 'package']:
        with self.argument_context(f'batch application {command}') as c:
            c.argument('account_name', batch_name_type, options_list=('--name', '-n'), validator=application_enabled)

    # TODO: Refactor so the help text can be extracted automatically
    with self.argument_context('batch pool resize') as c:
        c.argument('if_modified_since', help='The operation will be performed only if the resource has been modified since the specified timestamp.', type=datetime_format, arg_group='Pre-condition and Query')
        c.argument('if_unmodified_since', help='The operation will not be performed only if the resource has been modified since the specified timestamp.', type=datetime_format, arg_group='Pre-condition and Query')
        c.argument('if_match', help='The operation will be performed only if the resource\'s current ETag exactly matches the specified value.', arg_group='Pre-condition and Query')
        c.argument('if_none_match', help='The operation will not be performed only if the resource\'s current ETag exactly matches the specified value.', arg_group='Pre-condition and Query')
        c.argument('pool_id', help='The ID of the pool.')
        c.argument('abort', action='store_true', help='Stop the pool resize operation.', validator=validate_pool_resize_parameters)
        c.argument('node_deallocation_option', options_list=('--node-deallocation-option',), help='When nodes may be removed from the pool, if the pool size is decreasing.', arg_type=get_enum_type(BatchNodeDeallocationOption))
        c.argument('resize_timeout', type=duration_format, help='The default value is 15 minutes. The minimum value is 5 minutes. '
                                                                'If you specify a value less than 5 minutes, the Batch service '
                                                                'returns an error; if you are calling the REST API directly, the '
                                                                'HTTP status code is 400 (Bad Request). ISO-8601 duration format')
        c.argument('target_dedicated_nodes', help='The desired number of dedicated Compute Nodes in the Pool.')
        c.argument('target_low_priority_nodes', help='The desired number of Spot/Low-priority Compute Nodes in the Pool.')

    # TODO: Refactor so the help text can be extracted automatically
    with self.argument_context('batch pool reset') as c:
        c.argument('json_file', type=file_type, help='The file containing pool update properties parameter specification in JSON(formatted to match REST API request body). If this parameter is specified, all \'Pool Update Properties Parameter Arguments\' are ignored.', validator=validate_json_file, completer=FilesCompleter())
        c.argument('pool_id', help='The ID of the pool to update.')
        c.argument('application_package_references', nargs='+', type=batch_application_package_reference_format, arg_group='Pool', help='Required. The list replaces any existing Application Package '
                                                                                                                                        'references on the Pool. Changes to Application Package '
                                                                                                                                        'references affect all new Compute Nodes joining the Pool, '
                                                                                                                                        'but do not affect Compute Nodes that are already in the Pool '
                                                                                                                                        'until they are rebooted or reimaged. There is a maximum of '
                                                                                                                                        '10 Application Package references on any given Pool. If '
                                                                                                                                        'omitted, or if you specify an empty collection, any existing '
                                                                                                                                        'Application Packages references are removed from the Pool. A '
                                                                                                                                        'maximum of 10 references may be specified on a given Pool.')
        c.argument('metadata', nargs='+', type=metadata_item_format, arg_group='Pool', help='Required. This list replaces any existing metadata'
                                                                                            'configured on the Pool. If omitted, or if you specify an'
                                                                                            'empty collection, any existing metadata is removed from the'
                                                                                            'Pool.')
        c.argument('start_task_command_line', arg_group='Pool: Start Task',
                   help='The command line of the start task. The command line does not run under a shell, and therefore cannot take advantage of shell features such as environment variable expansion. If you want to take advantage of such features, you should invoke the shell in the command line, for example using "cmd /c MyCommand" in Windows or "/bin/sh -c MyCommand" in Linux.')
        c.argument('start_task_wait_for_success', action='store_true', arg_group='Pool: Start Task',
                   help='Whether the Batch service should wait for the start task to complete successfully (that is, to exit with exit code 0) before scheduling any tasks on the compute node. True if flag present, otherwise defaults to False.')
        c.argument('start_task_max_task_retry_count', arg_group='Pool: Start Task',
                   help='The maximum number of times the task may be retried.')
        c.argument('start_task_resource_files', nargs='+', type=resource_file_format, arg_group='Pool: Start Task',
                   help='A list of files that the Batch service will download to the '
                        'Compute Node before running the command line. Files listed under this element are '
                        'located in the Task\'s working directory. Space-separated '
                        'resource references in filename=httpurl format.')
        c.argument('start_task_environment_settings', nargs='+', type=environment_setting_format, arg_group='Pool: Start Task',
                   help='A list of environment variable settings for the start task. Space-separated values in \'key=value\' format.')

    with self.argument_context('batch pool autoscale enable') as c:
        c.extra('auto_scale_formula',
                help='The formula for the desired number of Compute Nodes in the Pool. '
                     'The formula is checked for validity before it is applied to the Pool. '
                     'If the formula is not valid, the Batch service rejects the request with detailed error information. '
                     'For more information about specifying this formula, see Automatically scale Compute Nodes in an Azure Batch Pool '
                     '(https://learn.microsoft.com/azure/batch/batch-automatic-scaling).')
        c.extra('auto_scale_evaluation_interval',
                help='The time interval at which to automatically adjust the Pool size according to the autoscale formula. '
                     'The default value is 15 minutes. The minimum and maximum value are 5 minutes and 168 hours respectively. '
                     'If you specify a value less than 5 minutes or greater than 168 hours, the Batch service rejects the request '
                     'with an invalid property value error; if you are calling the REST API directly, the HTTP status code is 400 (Bad Request). '
                     'If you specify a new interval, then the existing autoscale evaluation schedule will be stopped and a new autoscale evaluation '
                     'schedule will be started, with its starting time being the time when this request was issued.')

    with self.argument_context('batch private-endpoint-connection show') as c:
        c.argument('private_endpoint_connection_name', options_list=['--name', '-n'],
                   help='The private endpoint connection name. This must be unique within the account.')

    with self.argument_context('batch private-link-resource show') as c:
        c.argument('private_link_resource_name', options_list=['--name', '-n'],
                   help='The private link resource name. This must be unique within the account.')

    for command in [
        'pool delete', 'pool show', 'pool set', 'pool autoscale enable',
        'job delete', 'job show', 'job reset', 'job set', 'job disable', 'job enable', 'job stop',
        'job-schedule delete', 'job-schedule show', 'job-schedule reset', 'job-schedule set',
        'job-schedule disable', 'job-schedule enable', 'job-schedule stop',
        'task delete', 'task show', 'task stop', 'task reactivate',
        'node delete'
    ]:
        with self.argument_context(f'batch {command}') as c:
            c.extra('if_match', arg_group='Pre-condition and Query',
                    help='An ETag value associated with the version of the resource known to the client. '
                         'The operation will be performed only if the resource\'s current ETag on the service '
                         'exactly matches the value specified by the client.')
            c.extra('if_modified_since', arg_group='Pre-condition and Query', type=datetime_format,
                    help='A timestamp indicating the last modified time of the resource known to the client. '
                         'The operation will be performed only if the resource on the service has been modified '
                         'since the specified time.',)
            c.extra('if_none_match', arg_group='Pre-condition and Query',
                    help='An ETag value associated with the version of the resource known to the client. '
                         'The operation will be performed only if the resource\'s current ETag on the service '
                         'does not match the value specified by the client.')
            c.extra('if_unmodified_since', arg_group='Pre-condition and Query', type=datetime_format,
                    help='A timestamp indicating the last modified time of the resource known to the client. '
                         'The operation will be performed only if the resource on the service has been modified '
                         'since the specified time.')

    for command in ['pool node-counts list', 'pool supported-images list', 'job list', 'pool list', 'job-schedule list', 'task list', 'job prep-release-status list', 'node list', 'node file list', 'task file list', 'pool usage-metrics list']:
        with self.argument_context(f'batch {command}') as c:
            c.extra('filter', arg_group='Pre-condition and Query', help='An OData $filter clause. For more information on constructing this filter,'
                                                                        'see https://learn.microsoft.com/rest/api/batchservice/odata-filters-in-batch.')

    for command in ['job list', 'pool list', 'job-schedule list', 'task list', 'job show', 'pool show', 'task show', 'job-schedule show']:
        with self.argument_context(f'batch {command}') as c:
            c.extra('select', nargs='+', help='An OData $select clause.', arg_group='Pre-condition and Query')
            c.extra('expand', nargs='+', help='An OData $expand clause.', arg_group='Pre-condition and Query')

    with self.argument_context('batch job list') as c:
        c.argument('job_schedule_id', help='The ID of the job schedule from which you want to get a list of jobs. If omitted, lists all jobs in the account.')

    for command in ['node list', 'node show', 'job prep-release-status list', 'task subtask list']:
        with self.argument_context(f'batch {command}') as c:
            c.extra('select', nargs='+', help='An OData $select clause.', arg_group='Pre-condition and Query')

    with self.argument_context('batch job stop') as c:
        c.extra('reason', options_list=['--terminate-reason'], help='Termination reason. The text you want to appear as the job\'s TerminateReason. The default is \'UserTerminate\'.')

    for command in ['node file delete', 'task file delete']:
        with self.argument_context(f'batch {command}') as c:
            c.extra('recursive', arg_type=get_three_state_flag(),
                    help='Whether to delete children of a directory. If the filePath parameter '
                    'represents a directory instead of a file, you can set recursive to true to delete the '
                    'directory and all of the files and subdirectories in it. If recursive is false '
                    'then the directory must be empty or deletion will fail. Default value is None.')

    with self.argument_context('batch node file list') as c:
        c.extra('recursive', arg_type=get_three_state_flag(), help='Whether to list children of a directory.')

    with self.argument_context('batch task reset') as c:
        c.argument('if_modified_since', help='The operation will be performed only if the resource has been modified since the specified timestamp.', type=datetime_format, arg_group='Pre-condition and Query')
        c.argument('if_unmodified_since', help='The operation will not be performed only if the resource has been modified since the specified timestamp.', type=datetime_format, arg_group='Pre-condition and Query')
        c.argument('if_match', help='The operation will be performed only if the resource\'s current ETag exactly matches the specified value.', arg_group='Pre-condition and Query')
        c.argument('if_none_match', help='The operation will not be performed only if the resource\'s current ETag exactly matches the specified value.', arg_group='Pre-condition and Query')
        c.argument('job_id', help='The ID of the Job containing the Task.', required=True)
        c.argument('task_id', help='The ID of the Task to update.', required=True)
        c.argument('json_file', type=file_type, help='The file containing pool update properties parameter specification in JSON(formatted to match REST API request body). If this parameter is specified, all \'Pool Update Properties Parameter Arguments\' are ignored.', validator=validate_json_file, completer=FilesCompleter())
        c.argument('max_task_retry_count', arg_group='Constraints',
                   help='The maximum number of times the Task may be retried. The Batch service '
                        'retries a Task if its exit code is nonzero. Note that this value '
                        'specifically controls the number of retries for the Task executable due '
                        'to a nonzero exit code. The Batch service will try the Task once, and '
                        'may then retry up to this limit. For example, if the maximum retry '
                        'count is 3, Batch tries the Task up to 4 times (one initial try and 3 '
                        'retries). If the maximum retry count is 0, the Batch service does not '
                        'retry the Task after the first attempt. If the maximum retry count is '
                        '-1, the Batch service retries the Task without limit, however this is '
                        'not recommended for a start task or any task. The default value is 0 '
                        '(no retries).')
        c.argument('max_wall_clock_time', type=duration_format, arg_group='Constraints',
                   help='If this is not specified, there is no time limit on how long the Task '
                        'may run. Expected format is an ISO-8601 duration.')
        c.argument('retention_time', type=duration_format, arg_group='Constraints',
                   help='The default is 7 days, i.e. the Task directory will be retained for 7 '
                        'days unless the Compute Node is removed or the Job is deleted. Expected '
                        'format is an ISO-8601 duration.')

    with self.argument_context('batch pool usage-metrics list') as c:
        c.extra('endtime', options_list=['--end-time'], arg_group='Pre-condition and Query',
                help='The latest time from which to include metrics. This must be at least two '
                     'hours before the current time. If not specified this defaults to the end '
                     'time of the last aggregation interval currently available.')
        c.extra('starttime', options_list=['--start-time'], arg_group='Pre-condition and Query',
                help='The earliest time from which to include metrics. This must be at least two '
                     'and a half hours before the current time. If not specified this defaults to '
                     'the start time of the last aggregation interval currently available.')

    with self.argument_context('batch task file list') as c:
        c.extra('recursive', arg_type=get_three_state_flag(),
                help='Whether to list children of the Task directory. This parameter can be '
                'used in combination with the filter parameter to list specific type of files.')

    for command in ['node file download', 'task file download']:
        with self.argument_context(f'batch {command}') as c:
            c.extra('end_range', arg_group='Pre-condition and Query',
                    help='The byte range to be retrieved. If not set the file will be retrieved to the end.')
            c.extra('start_range', arg_group='Pre-condition and Query',
                    help='The byte range to be retrieved. If not set the file will be retrieved from the beginning.')

    for command in ['node file download', 'node file show', 'task file show', 'task file download']:
        with self.argument_context(f'batch {command}') as c:
            c.extra('if_modified_since', arg_group='Pre-condition and Query', type=datetime_format,
                    help='A timestamp indicating the last modified time of the resource known to the client. '
                         'The operation will be performed only if the resource on the service has been modified '
                         'since the specified time.',)
            c.extra('if_unmodified_since', arg_group='Pre-condition and Query', type=datetime_format,
                    help='A timestamp indicating the last modified time of the resource known to the client. '
                         'The operation will be performed only if the resource on the service has been modified '
                         'since the specified time.')

    with self.argument_context('batch pool create') as c:
        c.argument('json_file', help='The file containing pool create properties parameter specification in JSON(formatted to match REST API request body). If this parameter is specified, all \'Pool Create Properties Parameter Arguments\' are ignored.  See https://learn.microsoft.com/rest/api/batchservice/pools/create-pool?tabs=HTTP#request-body')
        c.argument('enable_accelerated_networking', arg_type=get_three_state_flag(), options_list=['--accelerated-networking'], arg_group="Pool: Network Configuration",
                   help='Whether this pool should enable accelerated networking. Accelerated networking enables single root I/O virtualization (SR-IOV) to a VM, which may lead to improved networking performance. For more details, see: https://learn.microsoft.com/azure/virtual-network/accelerated-networking-overview. Set true to enable.')
        c.argument('caching',
                   options_list=('--os-disk-caching'),
                   arg_type=get_enum_type(CachingType),
                   arg_group="Pool: OS Disk",
                   help="Specify the caching requirements. Possible values are: None, ReadOnly, ReadWrite. The default values are: None for Standard storage. ReadOnly for Premium storage.")
        c.argument('storage_account_type',
                   arg_group="Pool: OS Disk",
                   arg_type=get_enum_type(StorageAccountType),
                   help="The storage account type for managed disk")
        c.argument('disk_size_g_b',
                   options_list=('--os-disk-size'),
                   arg_group="Pool: OS Disk",
                   help="The initial disk size in GB when creating new OS disk.")
        c.argument('write_accelerator_enabled', arg_type=get_three_state_flag(),
                   options_list=('--enable-write-accel'),
                   arg_group="Pool: OS Disk",
                   help="Specify whether writeAccelerator should be enabled or disabled on the disk.")
        c.argument('encryption_at_host', arg_type=get_three_state_flag(),
                   arg_group='Pool: Security Profile',
                   help='This property can be used by user in the request to enable or disable the Host Encryption for the virtual machine or virtual machine scale set. This will enable the encryption for all the disks including Resource/Temp disk at host itself. The default behavior is: The Encryption at host will be disabled unless this property is set to true for the resource.')
        c.argument('security_type',
                   arg_group='Pool: Security Profile',
                   help='Specify the SecurityType of the virtual machine. It has to be set to any specified value to enable UefiSettings. The default behavior is: UefiSettings will not be enabled unless this property is set.')
        c.argument('secure_boot_enabled', arg_type=get_three_state_flag(),
                   options_list=('--enable-secure-boot'),
                   arg_group='Pool: Security Profile',
                   help='Enable secure boot')
        c.argument('v_tpm_enabled', arg_type=get_three_state_flag(),
                   options_list=('--enable-vtpm'),
                   arg_group='Pool: Security Profile',
                   help='Enable vTPM')
        c.argument('disable_automatic_rollback', options_list=['--disable-auto-rollback'], arg_type=get_three_state_flag())
        c.argument('enable_automatic_os_upgrade', options_list=['--enable-auto-os-upgrade'], arg_type=get_three_state_flag())
        c.argument('os_rolling_upgrade_deferral', options_list=['--defer-os-rolling-upgrade'], arg_type=get_three_state_flag())
        c.argument('use_rolling_upgrade_policy', arg_type=get_three_state_flag())
        c.extra('disk_encryption_targets',
                arg_group="Pool: Virtual Machine Configuration",
                help='A space separated list of DiskEncryptionTargets. current possible values include OsDisk and TemporaryDisk.', type=disk_encryption_configuration_format)
        c.extra('image', completer=load_supported_images, arg_group="Pool: Virtual Machine Configuration",
                help="OS image reference. This can be either 'publisher:offer:sku[:version]' format, or a fully qualified ARM image id of the form '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroup}/providers/Microsoft.Compute/images/{imageName}'. If 'publisher:offer:sku[:version]' format, version is optional and if omitted latest will be used. Valid values can be retrieved via 'az batch pool supported-images list'. For example: 'MicrosoftWindowsServer:WindowsServer:2012-R2-Datacenter:latest'")
        c.argument('mode', options_list=['--upgrade-policy-mode'], help='The mode of the pool OS upgrade.')
        c.argument('enable_cross_zone_upgrade', arg_type=get_three_state_flag())
        c.argument('prioritize_unhealthy_instances', arg_type=get_three_state_flag())
        c.argument('rollback_failed_instances_on_policy_breach', arg_type=get_three_state_flag())

    with self.argument_context('batch task create') as c:
        c.argument('json_file', type=file_type, help='The file containing the task(s) to create in JSON(formatted to match REST API request body). When submitting multiple tasks, accepts either an array of tasks or a TaskAddCollectionParamater. If this parameter is specified, all other parameters are ignored.', validator=validate_json_file, completer=FilesCompleter())
        c.argument('application_package_references', nargs='+', help='The space-separated list of IDs specifying the application packages to be installed. Space-separated application IDs with optional version in \'id[#version]\' format.', type=batch_application_package_reference_format)
        c.argument('job_id', help='The ID of the job containing the task.')
        c.argument('task_id', help='The ID of the task.')
        c.argument('command_line', help='The command line of the task. The command line does not run under a shell, and therefore cannot take advantage of shell features such as environment variable expansion. If you want to take advantage of such features, you should invoke the shell in the command line, for example using "cmd /c MyCommand" in Windows or "/bin/sh -c MyCommand" in Linux.')
        c.argument('environment_settings', nargs='+', help='A list of environment variable settings for the task. Space-separated values in \'key=value\' format.', type=environment_setting_format)
        c.argument('resource_files', nargs='+', help='A list of files that the Batch service will download to the compute node before running the command line. Space-separated resource references in filename=httpurl format, with httpurl being any HTTP url with public access or a SAS url with read access.', type=resource_file_format)
        c.argument('affinity_id', help='Required. You can pass the affinityId of a Node to indicate '
                                       'that this Task needs to run on that Compute Node. Note that '
                                       'this is just a soft affinity. If the target Compute Node is '
                                       'busy or unavailable at the time the Task is scheduled, then '
                                       'the Task will be scheduled elsewhere.')
        c.argument('max_task_retry_count', help='The maximum number of times the Task may be retried. The '
                                                'Batch service retries a Task if its exit code is nonzero. '
                                                'Note that this value specifically controls the number of '
                                                'retries for the Task executable due to a nonzero exit code. '
                                                'The Batch service will try the Task once, and may then retry '
                                                'up to this limit. For example, if the maximum retry count is '
                                                '3, Batch tries the Task up to 4 times (one initial try and 3 '
                                                'retries). If the maximum retry count is 0, the Batch service '
                                                'does not retry the Task after the first attempt. If the '
                                                'maximum retry count is -1, the Batch service retries the Task '
                                                'without limit, however this is not recommended for a start '
                                                'task or any task. The default value is 0 (no retries).')
        c.argument('max_wall_clock_time', type=duration_format, help='If this is not specified, there is no time limit on how long '
                                                                     'the Task may run.')
        c.argument('retention_time', type=duration_format, help='The default is 7 days, i.e. the Task directory will be '
                                                                'retained for 7 days unless the Compute Node is removed or the '
                                                                'Job is deleted.')

    for item in ['batch pool resize', 'batch pool reset', 'batch job list', 'batch task create', 'batch task reset']:
        with self.argument_context(item) as c:
            c.extra('account_name', arg_group='Batch Account', validator=validate_client_parameters,
                    help='The Batch account name. Only needed Alternatively, set by environment variable: AZURE_BATCH_ACCOUNT')
            c.extra('account_key', arg_group='Batch Account',
                    help='The Batch account key. Alternatively, set by environment variable: AZURE_BATCH_ACCESS_KEY')
            c.extra('account_endpoint', arg_group='Batch Account',
                    help='Batch service endpoint. Alternatively, set by environment variable: AZURE_BATCH_ENDPOINT')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from unittest.mock import Mock, patch
from azure.cli.command_modules.aro._dynamic_validators import (
    can_do_action,
    dyn_validate_cidr_ranges,
    dyn_validate_subnet_and_route_tables,
    dyn_validate_vnet,
    dyn_validate_resource_permissions,
    dyn_validate_version,
    dyn_validate_managed_identity_delete_permissions
)

from azure.mgmt.authorization.models import Permission
import pytest

test_can_do_action_data = [
    (
        "empty permissions list",
        [],
        "Microsoft.Network/virtualNetworks/subnets/join/action",
        False
    ),
    (
        "has permission - exact",
        [
            Permission(actions=["Microsoft.Compute/virtualMachines/*"], not_actions=[]),
            Permission(actions=["Microsoft.Network/virtualNetworks/subnets/join/action"], not_actions=[]),
        ],
        "Microsoft.Network/virtualNetworks/subnets/join/action",
        True
    ),
    (
        "has permission - wildcard",
        [
            Permission(actions=["Microsoft.Network/virtualNetworks/subnets/*/action"], not_actions=[]),
        ],
        "Microsoft.Network/virtualNetworks/subnets/join/action",
        True
    ),
    (
        "has permission - exact, conflict",
        [
            Permission(actions=[], not_actions=["Microsoft.Network/virtualNetworks/subnets/join/action"]),
            Permission(actions=["Microsoft.Network/virtualNetworks/subnets/join/action"], not_actions=[]),
        ],
        "Microsoft.Network/virtualNetworks/subnets/join/action",
        True
    ),
    (
        "has permission excluded - exact",
        [
            Permission(actions=["Microsoft.Network/*"], not_actions=["Microsoft.Network/virtualNetworks/subnets/join/action"]),
        ],
        "Microsoft.Network/virtualNetworks/subnets/join/action",
        False
    ),
    (
        "has permission excluded - wildcard",
        [
            Permission(actions=["Microsoft.Network/*"], not_actions=["Microsoft.Network/virtualNetworks/subnets/*/action"]),
        ],
        "Microsoft.Network/virtualNetworks/subnets/join/action",
        False
    )
]


@pytest.mark.parametrize(
    "test_description, perms, action, expected",
    test_can_do_action_data,
    ids=[i[0] for i in test_can_do_action_data]
)
def test_can_do_action(
    test_description, perms, action, expected
):
    actual = can_do_action(perms, action)

    if actual != expected:
        raise Exception(f"Error mismatch, expected: {expected}, actual: {actual}")


test_validate_cidr_data = [
    (
        "should return no error on address_prefix set on subnets, no additional cidrs, no overlap",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr=None, service_cidr=None),
        [{"addressPrefix": "172.143.5.0/24"}, {"addressPrefix": "172.143.4.0/25"}],
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        None
    ),
    (
        "should return no error on address_prefix set on subnets, additional cidrs, no overlap",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr="172.142.0.0/21", service_cidr="172.143.6.0/25"),
        [{"addressPrefix": "172.143.4.0/24"}, {"addressPrefix": "172.143.5.0/25"}],
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        None
    ),
    (
        "should return no error on multiple address_prefixes set on subnets, additional cidrs, no overlap",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr="172.142.0.0/21", service_cidr="172.143.6.0/25"),
        [{"addressPrefixes": ["172.143.4.0/24", "172.143.8.0/25"]}, {"addressPrefixes": ["172.143.5.0/25", "172.143.9.0/24"]}],
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        None
    ),
    (
        "should error on address_prefix set on subnets, no additional cidrs, overlap",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr=None, service_cidr=None),
        [{"addressPrefix": "172.143.4.0/24"}, {"addressPrefix": "172.143.4.0/25"}],
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        "172.143.4.0/24 is not valid as it overlaps with 172.143.4.0/25"
    ),
    (
        "should return error on pod cidr not having enough addresses to create cluster",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr="172.143.4.0/24", service_cidr=None),
        [{'addressPrefix': "172.143.5.0/24"}, {'addressPrefix': "172.143.4.0/25"}],
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        "172.143.4.0/24 does not contain enough addresses for 3 master nodes (Requires cidr prefix of 21 or lower)"
    ),
]


@pytest.mark.parametrize(
    "test_description, cmd_mock, namespace_mock, get_subnet_mock_side_effect, parse_resource_id_mock_return_value, expected_addresses",
    test_validate_cidr_data,
    ids=[i[0] for i in test_validate_cidr_data]
)
@patch('azure.cli.command_modules.aro._dynamic_validators.get_subnet')
@patch('azure.cli.command_modules.aro._dynamic_validators.parse_resource_id')
def test_validate_cidr(
    # Mocked functions:
    parse_resource_id_mock, get_subnet_mock,

    # Test cases parameters:
    test_description, cmd_mock, namespace_mock, get_subnet_mock_side_effect, parse_resource_id_mock_return_value, expected_addresses
):
    parse_resource_id_mock.return_value = parse_resource_id_mock_return_value
    get_subnet_mock.side_effect = get_subnet_mock_side_effect

    validate_cidr_fn = dyn_validate_cidr_ranges()
    if expected_addresses is None:
        addresses = validate_cidr_fn(cmd_mock, namespace_mock)

        if (len(addresses) > 0):
            raise Exception(f"Unexpected Error: {addresses[0]}")
    else:
        addresses = validate_cidr_fn(cmd_mock, namespace_mock)

        if (addresses[0][3] != expected_addresses):
            raise Exception(f"Error returned was not expected\n Expected : {expected_addresses}\n Actual   : {addresses[0][3]}")


test_validate_subnets_data = [
    (
        "should not return missing permission when actions are permitted",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr=None, service_cidr=None),
        {'routeTable': {'id': 'test'}},
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=["Microsoft.Network/routeTables/*"], not_actions=[])]}),
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        Mock(),
        None
    ),
    (
        "should return missing permission when actions are not permitted",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr=None, service_cidr=None),
        {'routeTable': {'id': 'test'}},
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=[], not_actions=["Microsoft.Network/routeTables/*"])]}),
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        Mock(),
        "Microsoft.Network/routeTables/join/action permission is missing"
    ),
    (
        "should return missing permission when actions are not present",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr=None, service_cidr=None),
        {'routeTable': {'id': 'test'}},
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=[], not_actions=[])]}),
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        Mock(),
        "Microsoft.Network/routeTables/join/action permission is missing"
    ),
    (
        "should return message when network security group is already attached to subnet",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr=None, service_cidr=None, enable_preconfigured_nsg=None),
        {'networkSecurityGroup': {'id': 'test'}, 'routeTable': {'id': 'test'}},
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=["Microsoft.Network/routeTables/*"], not_actions=[])]}),
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        Mock(),
        "A Network Security Group \"test\" is already assigned to this subnet. "
        "Ensure there are no Network Security Groups assigned to cluster "
        "subnets before cluster creation"
    ),
    (
        "should not return message when Preconfigured NSG is enabled and network security group is attached to subnet",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr=None, service_cidr=None, enable_preconfigured_nsg=True),
        {'networkSecurityGroup': {'id': 'test'}, 'routeTable': {'id': 'test'}},
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=["Microsoft.Network/routeTables/*"], not_actions=[])]}),
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        Mock(),
        None
    )
]


@pytest.mark.parametrize(
    "test_description, cmd_mock, namespace_mock, get_subnet_mock_return_value, auth_client_mock, parse_resource_id_mock_return_value, is_valid_resource_id_mock_return_value, expected_missing_perms",
    test_validate_subnets_data,
    ids=[i[0] for i in test_validate_subnets_data]
)
@patch('azure.cli.command_modules.aro._dynamic_validators.get_subnet')
@patch('azure.cli.command_modules.aro._dynamic_validators.get_mgmt_service_client')
@patch('azure.cli.command_modules.aro._dynamic_validators.parse_resource_id')
@patch('azure.cli.command_modules.aro._dynamic_validators.is_valid_resource_id')
def test_validate_subnets(
    # Mocked functions:
    is_valid_resource_id_mock, parse_resource_id_mock, get_mgmt_service_client_mock, get_subnet_mock,

    # Test cases parameters:
    test_description, cmd_mock, namespace_mock, get_subnet_mock_return_value, auth_client_mock, parse_resource_id_mock_return_value, is_valid_resource_id_mock_return_value, expected_missing_perms
):
    is_valid_resource_id_mock.return_value = is_valid_resource_id_mock_return_value
    parse_resource_id_mock.return_value = parse_resource_id_mock_return_value
    get_mgmt_service_client_mock.side_effect = [auth_client_mock]
    get_subnet_mock.return_value = get_subnet_mock_return_value

    validate_subnet_fn = dyn_validate_subnet_and_route_tables('')
    if expected_missing_perms is None:
        missing_perms = validate_subnet_fn(cmd_mock, namespace_mock)

        if (len(missing_perms) > 0):
            raise Exception(f"Unexpected Permission Missing: {missing_perms[0]}")
    else:
        missing_perms = validate_subnet_fn(cmd_mock, namespace_mock)

        if (missing_perms[0][3] != expected_missing_perms):
            raise Exception(f"Error returned was not expected\n Expected : {expected_missing_perms}\n Actual   : {missing_perms[0][3]}")


test_validate_vnets_data = [
    (
        "should not return missing permission when actions are permitted",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr=None, service_cidr=None),
        {'routeTable': {'id': 'test'}},
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=["Microsoft.Network/virtualNetworks/*"], not_actions=[])]}),
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        Mock(),
        None
    ),
    (
        "should return disabled permission when actions are not permitted",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr=None, service_cidr=None),
        {'routeTable': {'id': 'test'}},
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=[], not_actions=["Microsoft.Network/virtualNetworks/*"])]}),
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        Mock(),
        "Microsoft.Network/virtualNetworks/join/action permission is missing"
    ),
    (
        "should return missing permission when actions are not present",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(vnet='', key="192.168.0.1/32", master_subnet='', worker_subnet='', pod_cidr=None, service_cidr=None),
        {'routeTable': {'id': 'test'}},
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=[], not_actions=[])]}),
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        Mock(),
        "Microsoft.Network/virtualNetworks/join/action permission is missing"
    )
]


@pytest.mark.parametrize(
    "test_description, cmd_mock, namespace_mock, get_vnet_mock_return_value, auth_client_mock, parse_resource_id_mock_return_value, is_valid_resource_id_mock_return_value, expected_missing_perms",
    test_validate_vnets_data,
    ids=[i[0] for i in test_validate_vnets_data]
)
@patch('azure.cli.command_modules.aro._dynamic_validators.get_vnet')
@patch('azure.cli.command_modules.aro._dynamic_validators.get_mgmt_service_client')
@patch('azure.cli.command_modules.aro._dynamic_validators.parse_resource_id')
@patch('azure.cli.command_modules.aro._dynamic_validators.is_valid_resource_id')
def test_validate_vnets(
    # Mocked functions:
    is_valid_resource_id_mock, parse_resource_id_mock, get_mgmt_service_client_mock, get_vnet_mock,

    # Test cases parameters:
    test_description, cmd_mock, namespace_mock, get_vnet_mock_return_value, auth_client_mock, parse_resource_id_mock_return_value, is_valid_resource_id_mock_return_value, expected_missing_perms
):
    is_valid_resource_id_mock.return_value = is_valid_resource_id_mock_return_value
    parse_resource_id_mock.return_value = parse_resource_id_mock_return_value
    get_mgmt_service_client_mock.side_effect = [auth_client_mock]
    get_vnet_mock.return_value = get_vnet_mock_return_value

    validate_vnet_fn = dyn_validate_vnet("vnet")
    if expected_missing_perms is None:
        missing_perms = validate_vnet_fn(cmd_mock, namespace_mock)

        if (len(missing_perms) > 0):
            raise Exception(f"Unexpected Permission Missing: {missing_perms[0]}")
    else:
        missing_perms = validate_vnet_fn(cmd_mock, namespace_mock)

        if (missing_perms[0][3] != expected_missing_perms):
            raise Exception(f"Error returned was not expected\n Expected : {expected_missing_perms}\n Actual   : {missing_perms[0][3]}")


test_validate_resource_data = [
    (
        "should not return missing permission when role assignments are assigned",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(),
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": None,
            "child_name_1": None
        },
        [True, True, True, True, True, True, True, True],
        None
    ),
    (
        "should return missing permission when role assignments are not assigned",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(),
        {
            "subscription": "subscription",
            "namespace": "MICROSOFT.NETWORK",
            "type": "virtualnetworks",
            "last_child_num": 1,
            "child_type_1": "subnets",
            "resource_group": None,
            "name": "Test_Subnet",
            "child_name_1": None
        },
        [True, True, True, True, True, True, True, False],
        "Resource Test_Subnet is missing role assignment test for service principal test " +
        "(These roles will be automatically added during cluster creation)"
    )
]


@pytest.mark.parametrize(
    "test_description, cmd_mock, namespace_mock, parse_resource_id_mock_return_value, has_role_assignment_on_resource_mock_return_value, expected_missing_perms",
    test_validate_resource_data,
    ids=[i[0] for i in test_validate_resource_data]
)
@patch('azure.cli.command_modules.aro._dynamic_validators.has_role_assignment_on_resource')
@patch('azure.cli.command_modules.aro._dynamic_validators.parse_resource_id')
def test_validate_resources(
    # Mocked functions:
    parse_resource_id_mock, has_role_assignment_on_resource_mock,

    # Test cases parameters:
    test_description, cmd_mock, namespace_mock, parse_resource_id_mock_return_value, has_role_assignment_on_resource_mock_return_value, expected_missing_perms
):
    parse_resource_id_mock.return_value = parse_resource_id_mock_return_value
    has_role_assignment_on_resource_mock.side_effect = has_role_assignment_on_resource_mock_return_value

    sp_ids = ["test", "test"]
    resources = {"test": "test", "test": "test"}
    validate_res_perms_fn = dyn_validate_resource_permissions(sp_ids, resources)
    if expected_missing_perms is None:
        missing_perms = validate_res_perms_fn(cmd_mock, namespace_mock)

        if (len(missing_perms) > 0):
            raise Exception(f"Unexpected Permission Missing: {missing_perms[0]}")
    else:
        missing_perms = validate_res_perms_fn(cmd_mock, namespace_mock)

        if (missing_perms[0][3] != expected_missing_perms):
            raise Exception(f"Error returned was not expected\n Expected : {expected_missing_perms}\n Actual   : {missing_perms[0][3]}")


test_validate_version_data = [
    (
        "should not return error when visibility is Public",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(version="4.9.10"),
        ["4.9.10"],
        None
    ),
    (
        "should return error when visibility is random string",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(version="72.2343.21212"),
        ["4.9.10"],
        "72.2343.21212 is not a valid version, valid versions are ['4.9.10']"
    )
]


@pytest.mark.parametrize(
    "test_description, cmd_mock, namespace_mock, get_versions_mock_return_value, expected_errors",
    test_validate_version_data,
    ids=[i[0] for i in test_validate_version_data]
)
@patch('azure.cli.command_modules.aro.custom.aro_get_versions')
def test_validate_version(

    # Mocked Functions
    aro_get_versions_mock,
    # Test cases parameters:
    test_description, cmd_mock, namespace_mock, get_versions_mock_return_value, expected_errors
):
    aro_get_versions_mock.return_value = get_versions_mock_return_value

    validate_version_fn = dyn_validate_version()
    if expected_errors is None:
        errors = validate_version_fn(cmd_mock, namespace_mock)

        if (len(errors) > 0):
            raise Exception(f"Unexpected Error: {errors[0][3]}")
    else:
        errors = validate_version_fn(cmd_mock, namespace_mock)

        if (errors[0][3] != expected_errors):
            raise Exception(f"Error returned was not expected\n Expected : {expected_errors}\n Actual   : {errors[0][3]}")


test_dyn_validate_managed_identity_delete_permissions_data = [
    (
        "should not return missing permissions when actions are permitted",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(managed_identities=["/subscriptions/sub/resourceGroups/rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/identity1"]),
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=["Microsoft.ManagedIdentity/userAssignedIdentities/delete"], not_actions=[])]}),
        {
            "resource_group": "rg",
            "name": "identity1",
            "namespace": "Microsoft.ManagedIdentity",
            "type": "userAssignedIdentities"
        },
        None
    ),
    (
        "should return missing permissions when actions are not permitted",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(managed_identities=["/subscriptions/sub/resourceGroups/rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/identity1"]),
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=[], not_actions=["Microsoft.ManagedIdentity/userAssignedIdentities/delete"])]}),
        {
            "resource_group": "rg",
            "name": "identity1",
            "namespace": "Microsoft.ManagedIdentity",
            "type": "userAssignedIdentities"
        },
        "Microsoft.ManagedIdentity/userAssignedIdentities/delete permission is missing over /subscriptions/sub/resourceGroups/rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/identity1"
    ),
    (
        "should return missing permissions when actions are missing",
        Mock(cli_ctx=Mock(get_progress_controller=Mock(add=Mock(), end=Mock()))),
        Mock(managed_identities=["/subscriptions/sub/resourceGroups/rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/identity1"]),
        Mock(**{"permissions.list_for_resource.return_value": [Permission(actions=[], not_actions=[])]}),
        {
            "resource_group": "rg",
            "name": "identity1",
            "namespace": "Microsoft.ManagedIdentity",
            "type": "userAssignedIdentities"
        },
        "Microsoft.ManagedIdentity/userAssignedIdentities/delete permission is missing over /subscriptions/sub/resourceGroups/rg/providers/Microsoft.ManagedIdentity/userAssignedIdentities/identity1"
    )
]


@pytest.mark.parametrize(
    "test_description, cmd_mock, namespace_mock, auth_client_mock, parse_resource_id_mock_return_value, expected_missing_perms",
    test_dyn_validate_managed_identity_delete_permissions_data,
    ids=[i[0] for i in test_dyn_validate_managed_identity_delete_permissions_data]
)
@patch("azure.cli.command_modules.aro._dynamic_validators.get_mgmt_service_client")
@patch("azure.cli.command_modules.aro._dynamic_validators.parse_resource_id")
def test_dyn_validate_managed_identity_delete_permissions(
    # Mocked functions:
    parse_resource_id_mock, get_mgmt_service_client_mock,
    # Test cases parameters:
    test_description, cmd_mock, namespace_mock, auth_client_mock, parse_resource_id_mock_return_value, expected_missing_perms
):
    parse_resource_id_mock.return_value = parse_resource_id_mock_return_value
    get_mgmt_service_client_mock.side_effect = [auth_client_mock]

    validate_fn = dyn_validate_managed_identity_delete_permissions()
    missing_perms = validate_fn(cmd_mock, namespace_mock)

    if expected_missing_perms is None:
        assert len(missing_perms) == 0, f"Unexpected errors: {missing_perms}"
    else:
        assert len(missing_perms) > 0, "Expected errors but got none"
        assert expected_missing_perms in missing_perms[0], f"Expected error '{expected_missing_perms}' but got '{missing_perms[0]}'"



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.log import get_logger
from azure.cli.core.util import CLIError

from ._utils import (
    get_resource_group_name_by_registry_name,
    validate_premium_registry
)

logger = get_logger(__name__)


REPLICATIONS_NOT_SUPPORTED = 'Replications are only supported for managed registries in Premium SKU.'


def acr_replication_list(cmd, client, registry_name, resource_group_name=None):
    _, resource_group_name = validate_premium_registry(
        cmd, registry_name, resource_group_name, REPLICATIONS_NOT_SUPPORTED)
    return client.list(resource_group_name, registry_name)


def acr_replication_create(cmd,
                           client,
                           location,
                           registry_name,
                           resource_group_name=None,
                           replication_name=None,
                           global_endpoint_routing=None,
                           zone_redundancy=None,
                           tags=None):
    registry, resource_group_name = validate_premium_registry(
        cmd, registry_name, resource_group_name, REPLICATIONS_NOT_SUPPORTED)

    normalized_location = "".join(location.split()).lower()
    if registry.location == normalized_location:
        raise CLIError('Replication could not be created in the same location as the registry.')

    from msrest.exceptions import ValidationError
    ReplicationType = cmd.get_models('Replication')

    replication_name = replication_name or normalized_location
    replication_properties = ReplicationType(
        location=location,
        region_endpoint_enabled=global_endpoint_routing,
        zone_redundancy=zone_redundancy,
        tags=tags)

    try:
        return client.begin_create(
            resource_group_name=resource_group_name,
            registry_name=registry_name,
            replication_name=replication_name,
            replication=replication_properties
        )
    except ValidationError as e:
        raise CLIError(e)


def acr_replication_delete(cmd,
                           client,
                           replication_name,
                           registry_name,
                           resource_group_name=None):
    _, resource_group_name = validate_premium_registry(
        cmd, registry_name, resource_group_name, REPLICATIONS_NOT_SUPPORTED)
    return client.begin_delete(resource_group_name, registry_name, replication_name)


def acr_replication_show(cmd,
                         client,
                         replication_name,
                         registry_name,
                         resource_group_name=None):
    _, resource_group_name = validate_premium_registry(
        cmd, registry_name, resource_group_name, REPLICATIONS_NOT_SUPPORTED)
    return client.get(resource_group_name, registry_name, replication_name)


def acr_replication_update_custom(instance, global_endpoint_routing=None, tags=None):
    if tags is not None:
        instance.tags = tags

    if global_endpoint_routing is not None:
        instance.region_endpoint_enabled = global_endpoint_routing

    return instance


def acr_replication_update_get(cmd):
    """Returns an empty ReplicationUpdateParameters object.
    """
    ReplicationUpdateParameters = cmd.get_models('ReplicationUpdateParameters')
    return ReplicationUpdateParameters()


def acr_replication_update_set(cmd,
                               client,
                               replication_name,
                               registry_name,
                               resource_group_name=None,
                               parameters=None):

    resource_group_name = get_resource_group_name_by_registry_name(
        cmd.cli_ctx, registry_name, resource_group_name)
    ReplicationUpdateParameters = cmd.get_models('ReplicationUpdateParameters')
    replication_update_parameters = ReplicationUpdateParameters(
        region_endpoint_enabled=parameters.region_endpoint_enabled,
        tags=parameters.tags
    )
    return client.begin_update(
        resource_group_name=resource_group_name,
        registry_name=registry_name,
        replication_name=replication_name,
        replication_update_parameters=replication_update_parameters)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Context-enriched error builder for az webapp deploy / az webapp up.
Enabled via the --enriched-errors flag on az webapp deploy / az webapp up.
"""

import re

from knack.log import get_logger
from knack.util import CLIError

from ._deployment_failure_patterns import match_failure_pattern

logger = get_logger(__name__)


class EnrichedDeploymentError(CLIError):
    # A CLIError subclass for context-enriched deployment failures.
    pass


_STATUS_CODE_PATTERNS = [
    re.compile(r'Status\s*Code[:\s]+(\d{3})', re.IGNORECASE),   # "Status Code: 400"
    re.compile(r'\(([45]\d{2})\)'),                             # "Bad Request(400)"
    re.compile(r'HTTP\s+(\d{3})', re.IGNORECASE),                # "HTTP 504"
    re.compile(
        r'\b([45]\d{2})\s+(?:Bad|Unauthorized|Forbidden|Not\s+Found|Conflict'
        r'|Too\s+Many|Internal|Gateway|Service)', re.IGNORECASE),  # "400 Bad Request"
]


def extract_status_code_from_message(message):
    if not message:
        return None
    for pattern in _STATUS_CODE_PATTERNS:
        m = pattern.search(message)
        if m:
            code = int(m.group(1))
            if 400 <= code <= 599:
                return code
    return None


def _get_app_runtime(cmd, resource_group_name, webapp_name, slot=None):
    try:
        from ._client_factory import web_client_factory
        client = web_client_factory(cmd.cli_ctx)
        if slot:
            config = client.web_apps.get_configuration_slot(resource_group_name, webapp_name, slot)
        else:
            config = client.web_apps.get_configuration(resource_group_name, webapp_name)
        if config.linux_fx_version:
            return config.linux_fx_version
        return "Unknown"
    except Exception:  # pylint: disable=broad-except
        return "Unknown"


def _get_app_region_and_plan_sku(cmd, resource_group_name, webapp_name):
    try:
        from ._client_factory import web_client_factory
        from azure.mgmt.core.tools import parse_resource_id
        from .utils import get_site_server_farm_id
        client = web_client_factory(cmd.cli_ctx)
        app = client.web_apps.get(resource_group_name, webapp_name)
        region = app.location if app else "Unknown"
        sku = "Unknown"
        server_farm_id = get_site_server_farm_id(app) if app else None
        if app and server_farm_id:
            plan_parts = parse_resource_id(server_farm_id)
            plan = client.app_service_plans.get(plan_parts['resource_group'], plan_parts['name'])
            if plan and plan.sku:
                sku = plan.sku.name
        return region, sku
    except Exception:  # pylint: disable=broad-except
        return "Unknown", "Unknown"


_ARTIFACT_TYPE_MAP = {
    'zip': 'ZipDeploy', 'war': 'WarDeploy', 'jar': 'JarDeploy',
    'ear': 'EarDeploy', 'startup': 'StartupFile', 'static': 'StaticDeploy'
}


def _determine_deployment_type(params=None, *, src_url=None, artifact_type=None):
    _src_url = src_url if src_url is not None else (getattr(params, 'src_url', None) if params else None)
    _artifact = artifact_type if artifact_type is not None else (
        getattr(params, 'artifact_type', None) if params else None)

    if _src_url:
        return "OneDeploy (URL-based)"

    return _ARTIFACT_TYPE_MAP.get(_artifact, "OneDeploy")


def build_enriched_error_context(params=None, *, cmd=None, resource_group_name=None,  # pylint: disable=too-many-locals
                                 webapp_name=None, slot=None, src_url=None,
                                 artifact_type=None, status_code=None, error_message=None,
                                 deployment_status=None,
                                 last_known_step=None, kudu_status=None):
    _cmd = cmd or (params.cmd if params else None)
    _rg = resource_group_name or (params.resource_group_name if params else None)
    _name = webapp_name or (params.webapp_name if params else None)
    _slot = slot if slot is not None else (
        getattr(params, 'slot', None) if params else None)
    _src_url = src_url if src_url is not None else (
        getattr(params, 'src_url', None) if params else None)
    _artifact = artifact_type if artifact_type is not None else (
        getattr(params, 'artifact_type', None) if params else None)

    pattern = match_failure_pattern(
        status_code=status_code,
        error_message=error_message,
    )

    # Build base context
    context = {}

    if pattern:
        context["errorCode"] = pattern["errorCode"]
        context["stage"] = pattern["stage"]
    else:
        context["errorCode"] = f"HTTP_{status_code}" if status_code else "UnknownDeploymentError"
        context["stage"] = deployment_status or "Unknown"

    # App metadata (best-effort)
    if _cmd and _rg and _name:
        context["runtime"] = _get_app_runtime(_cmd, _rg, _name, _slot)
        region, plan_sku = _get_app_region_and_plan_sku(_cmd, _rg, _name)
        context["region"] = region
        context["planSku"] = plan_sku
    else:
        context["runtime"] = "Unknown"
        context["region"] = "Unknown"
        context["planSku"] = "Unknown"

    context["deploymentType"] = _determine_deployment_type(
        params, src_url=_src_url, artifact_type=_artifact
    )

    # Suggested fixes
    if pattern:
        context["suggestedFixes"] = pattern["suggestedFixes"]
    else:
        context["suggestedFixes"] = [
            "Check deployment logs: 'az webapp log deployment show -n {} -g {}'".format(
                _name or '<app>', _rg or '<rg>'),
            "Check runtime logs: 'az webapp log tail -n {} -g {}'".format(
                _name or '<app>', _rg or '<rg>')
        ]

    # Extra diagnostics
    if last_known_step:
        context["lastKnownStep"] = last_known_step
    if kudu_status:
        context["kuduStatus"] = str(kudu_status)

    # Raw details
    if error_message:
        if len(error_message) > 500:
            context["rawError"] = error_message[:500] + "... [truncated]"
        else:
            context["rawError"] = error_message

    return context


def format_enriched_error_message(context):
    lines = []
    lines.append("")
    lines.append("=" * 72)
    lines.append("DEPLOYMENT FAILED: Context-Enriched Diagnostics")
    lines.append("=" * 72)
    lines.append("")

    lines.append(f"Error Code  : {context.get('errorCode', 'Unknown')}")
    lines.append(f"Stage       : {context.get('stage', 'Unknown')}")
    lines.append(f"Runtime     : {context.get('runtime', 'Unknown')}")
    lines.append(f"Deploy Type : {context.get('deploymentType', 'Unknown')}")
    lines.append(f"Region      : {context.get('region', 'Unknown')}")
    lines.append(f"Plan SKU    : {context.get('planSku', 'Unknown')}")
    if context.get("lastKnownStep"):
        lines.append(f"Last Step   : {context['lastKnownStep']}")
    if context.get("kuduStatus"):
        lines.append(f"Kudu Status : {context['kuduStatus']}")
    lines.append("")

    if context.get("rawError"):
        lines.append(f"Raw Error   : {context['rawError']}")
        lines.append("")

    fixes = context.get("suggestedFixes", [])
    if fixes:
        lines.append("Suggested Fixes:")
        for f in fixes:
            lines.append(f"  - {f}")
        lines.append("")

    # Copilot prompt
    lines.append("-" * 72)
    lines.append("  Copy the full error output above and paste it into GitHub Copilot Chat")
    lines.append("  with the prompt: 'Why did my Linux App Service deployment fail and how do I fix it?'")
    lines.append("-" * 72)

    return "\n".join(lines)


def raise_enriched_deployment_error(params=None, *, cmd=None, resource_group_name=None,
                                    webapp_name=None, slot=None, src_url=None,
                                    artifact_type=None, status_code=None, error_message=None,
                                    deployment_status=None,
                                    last_known_step=None, kudu_status=None):
    context = build_enriched_error_context(
        params=params,
        cmd=cmd,
        resource_group_name=resource_group_name,
        webapp_name=webapp_name,
        slot=slot,
        src_url=src_url,
        artifact_type=artifact_type,
        status_code=status_code,
        error_message=error_message,
        deployment_status=deployment_status,
        last_known_step=last_known_step,
        kudu_status=kudu_status
    )

    logger.debug("Deployment failure context: %s", context)

    message = format_enriched_error_message(context)
    raise EnrichedDeploymentError(message)


def build_enriched_plan_error_context(*, resource_group_name=None, plan_name=None,
                                      location=None, sku=None, status_code=None,
                                      error_message=None, last_known_step=None):
    from ._deployment_failure_patterns import match_control_plane_failure_pattern

    pattern = match_control_plane_failure_pattern(
        status_code=status_code,
        error_message=error_message,
    )

    context = {}

    if pattern:
        context["errorCode"] = pattern["errorCode"]
        context["stage"] = pattern["stage"]
        context["suggestedFixes"] = pattern["suggestedFixes"]
    else:
        context["errorCode"] = f"HTTP_{status_code}" if status_code else "UnknownPlanCreateError"
        context["stage"] = "ResourceProvisioning"
        context["suggestedFixes"] = [
            "Review the raw error message below for the failing property",
            "Verify --sku and --location are valid: 'az appservice list-locations --sku <SKU>'",
            "Confirm the resource group exists and you have Contributor access"
        ]

    context["resourceGroup"] = resource_group_name or "Unknown"
    context["planName"] = plan_name or "Unknown"
    context["region"] = location or "Unknown"
    context["planSku"] = sku or "Unknown"

    if last_known_step:
        context["lastKnownStep"] = last_known_step

    if error_message:
        if len(error_message) > 500:
            context["rawError"] = error_message[:500] + "... [truncated]"
        else:
            context["rawError"] = error_message

    return context


def format_enriched_plan_error_message(context):
    lines = []
    lines.append("")
    lines.append("=" * 72)
    lines.append("APP SERVICE PLAN CREATION FAILED: Context-Enriched Diagnostics")
    lines.append("=" * 72)
    lines.append("")

    lines.append(f"Error Code  : {context.get('errorCode', 'Unknown')}")
    lines.append(f"Stage       : {context.get('stage', 'Unknown')}")
    lines.append(f"Plan Name   : {context.get('planName', 'Unknown')}")
    lines.append(f"Resource Grp: {context.get('resourceGroup', 'Unknown')}")
    lines.append(f"Region      : {context.get('region', 'Unknown')}")
    lines.append(f"Plan SKU    : {context.get('planSku', 'Unknown')}")
    if context.get("lastKnownStep"):
        lines.append(f"Last Step   : {context['lastKnownStep']}")
    lines.append("")

    if context.get("rawError"):
        lines.append(f"Raw Error   : {context['rawError']}")
        lines.append("")

    fixes = context.get("suggestedFixes", [])
    if fixes:
        lines.append("Suggested Fixes:")
        for f in fixes:
            lines.append(f"  - {f}")
        lines.append("")

    # Copilot prompt
    lines.append("-" * 72)
    lines.append("  Copy the full error output above and paste it into GitHub Copilot Chat")
    lines.append("  with the prompt: 'Why did my az appservice plan create fail and how do I fix it?'")
    lines.append("-" * 72)

    return "\n".join(lines)


def raise_enriched_plan_error(*, resource_group_name=None, plan_name=None,
                              location=None, sku=None, status_code=None,
                              error_message=None, last_known_step=None):
    context = build_enriched_plan_error_context(
        resource_group_name=resource_group_name,
        plan_name=plan_name,
        location=location,
        sku=sku,
        status_code=status_code,
        error_message=error_message,
        last_known_step=last_known_step,
    )

    logger.debug("App Service plan creation failure context: %s", context)

    message = format_enriched_plan_error_message(context)
    raise EnrichedDeploymentError(message)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=protected-access, too-few-public-methods

import copy
import os
from argparse import Action
from collections import OrderedDict

from knack.log import get_logger
from knack.prompting import NoTTYException

from azure.cli.core import azclierror
from ._base import AAZUndefined, AAZBlankArgValue
from ._help import AAZShowHelp
from ._utils import AAZShortHandSyntaxParser
from ._prompt import AAZPromptInput
from .exceptions import AAZInvalidShorthandSyntaxError, AAZInvalidValueError, AAZUnknownFieldError

logger = get_logger(__name__)

_ELEMENT_APPEND_KEY = "_+_"  # used for list append


class AAZPromptInputOperation:

    def __init__(self, prompt: AAZPromptInput, action_cls):
        self.prompt = prompt
        self.action_cls = action_cls

    def __call__(self, *args, **kwargs):
        try:
            data = self.prompt()
        except NoTTYException:
            raise AAZInvalidValueError("argument value cannot be blank in non-interactive mode.")
        try:
            return self.action_cls.format_data(data)
        except (ValueError, KeyError) as ex:
            raise azclierror.InvalidArgumentValueError(f"Failed to parse value: {ex}") from ex


class AAZArgActionOperations:
    """AAZArg operations container"""

    def __init__(self):
        self._ops = []

    def add(self, data, *keys):
        self._ops.append((keys, data))

    def apply(self, args, dest):
        for keys, data in self._ops:
            if isinstance(data, AAZPromptInputOperation):
                data = data()
            self._assign_data(args, data, dest, *keys)

    @staticmethod
    def _assign_data(args, data, dest, *keys):
        from ._field_value import AAZList
        arg = args
        key = dest
        i = 0
        while i < len(keys):
            arg = arg[key]
            key = keys[i]
            if key == _ELEMENT_APPEND_KEY:
                assert isinstance(arg, AAZList)
                key = len(arg)
            i += 1
        arg[key] = data


class AAZArgAction(Action):
    _schema = None

    _str_parser = AAZShortHandSyntaxParser()  # support to parse normal string or short hand syntax string

    def __call__(self, parser, namespace, values, option_string=None):
        if not isinstance(getattr(namespace, self.dest), AAZArgActionOperations):
            # overwrite existing namespace value which is not an instance of AAZArgActionOperations.
            # especially the default value of argument.
            setattr(namespace, self.dest, AAZArgActionOperations())
        dest_ops = getattr(namespace, self.dest)
        try:
            try:
                self.setup_operations(dest_ops, values)
            except AAZShowHelp as aaz_help:
                # show help message
                aaz_help.keys = (option_string, *aaz_help.keys)
                self.show_aaz_help(parser, aaz_help)  # may raise AAZUnknownFieldError
        except (ValueError, KeyError) as ex:
            raise azclierror.InvalidArgumentValueError(f"Failed to parse '{option_string}' argument: {ex}") from ex

    @classmethod
    def setup_operations(cls, dest_ops, values, prefix_keys=None):
        raise NotImplementedError()

    @classmethod
    def format_data(cls, data):
        """format input data"""
        raise NotImplementedError()

    @staticmethod
    def show_aaz_help(parser, aaz_help):
        aaz_help.show()
        parser.exit()


class AAZSimpleTypeArgAction(AAZArgAction):

    @classmethod
    def setup_operations(cls, dest_ops, values, prefix_keys=None):
        if prefix_keys is None:
            prefix_keys = []
        if values is None:
            data = AAZBlankArgValue  # use blank data when values string is None
        else:
            if isinstance(values, list):
                assert prefix_keys  # the values will be input as an list when parse singular option of a list argument
                if len(values) != 1:
                    raise AAZInvalidValueError(f"only support 1 value, got {len(values)}: {values}")
                values = values[0]

            if isinstance(values, str) and len(values) > 0:
                try:
                    data = cls.decode_str(values)
                except AAZShowHelp as aaz_help:
                    aaz_help.schema = cls._schema
                    raise aaz_help
            else:
                data = values
        data = cls.format_data(data)
        dest_ops.add(data, *prefix_keys)

    @classmethod
    def decode_str(cls, value):
        return cls._str_parser(value, is_simple=True)

    @classmethod
    def format_data(cls, data):
        if data == AAZBlankArgValue:
            if cls._schema._blank == AAZUndefined:
                raise AAZInvalidValueError("argument value cannot be blank")
            if isinstance(cls._schema._blank, AAZPromptInput):
                # Postpone the prompt input when apply the operation.
                # In order not to break the logic of displaying help or validating other parameters.
                return AAZPromptInputOperation(prompt=cls._schema._blank, action_cls=cls)
            data = copy.deepcopy(cls._schema._blank)

        if data is None:
            if cls._schema._nullable:
                return data
            raise AAZInvalidValueError("field is not nullable")

        if cls._schema.DataType == str and isinstance(data, (int, bool, float)):
            data = str(data).lower()  # convert to json raw values

        if isinstance(data, str):
            # transfer string into correct data
            if cls._schema.enum:
                return cls._schema.enum[data]
            return cls._schema.DataType(data)   # str, int, float, boolean

        if isinstance(data, cls._schema.DataType):
            return data

        if isinstance(data, int) and cls._schema.DataType == float:
            return data

        raise AAZInvalidValueError(f"{cls._schema.DataType} type value expected, got '{data}'({type(data)})")


class AAZAnyTypeArgAction(AAZArgAction):

    @classmethod
    def setup_operations(cls, dest_ops, values, prefix_keys=None):
        if prefix_keys is None:
            prefix_keys = []
        if values is None:
            data = AAZBlankArgValue  # use blank data when values string is None
        else:
            if isinstance(values, list):
                assert prefix_keys  # the values will be input as an list when parse singular option of a list argument
                if len(values) != 1:
                    raise AAZInvalidValueError(f"only support 1 value, got {len(values)}: {values}")
                values = values[0]

            if isinstance(values, str) and len(values) > 0:
                try:
                    data = cls.decode_str(values)
                except AAZShowHelp as aaz_help:
                    aaz_help.schema = cls._schema
                    raise aaz_help
            else:
                data = values
        data = cls.format_data(data)
        dest_ops.add(data, *prefix_keys)

    @classmethod
    def decode_str(cls, value):
        from azure.cli.core.util import get_file_json, shell_safe_json_parse, get_file_yaml

        # check if the value is a partial value
        key, _, v = cls._str_parser.split_partial_value(value)
        if key is not None:
            raise AAZInvalidValueError(
                "AnyType args only support full value shorthand syntax, "
                "please don't use partial value shorthand syntax. "
                "If it's a simple string, please wrap it with single quotes.")

        # read from file
        path = os.path.expanduser(value)
        if os.path.exists(path):
            if path.endswith('.yml') or path.endswith('.yaml'):
                # read from yaml file
                v = get_file_yaml(path)
            else:
                # read from json file
                v = get_file_json(path, preserve_order=True)
        else:
            try:
                v = cls._str_parser(value)
            except AAZInvalidShorthandSyntaxError as shorthand_ex:
                try:
                    v = shell_safe_json_parse(value, True)
                except Exception as ex:
                    logger.debug(ex)  # log parse json failed expression
                    raise shorthand_ex  # raise shorthand syntax exception
        return v

    @classmethod
    def format_data(cls, data):
        if data == AAZBlankArgValue:
            if cls._schema._blank == AAZUndefined:
                raise AAZInvalidValueError("argument value cannot be blank")
            if isinstance(cls._schema._blank, AAZPromptInput):
                # Postpone the prompt input when apply the operation.
                # In order not to break the logic of displaying help or validating other parameters.
                return AAZPromptInputOperation(prompt=cls._schema._blank, action_cls=cls)
            data = copy.deepcopy(cls._schema._blank)

        if data is None:
            if cls._schema._nullable:
                return data
            raise AAZInvalidValueError("field is not nullable")
        return data


class AAZCompoundTypeArgAction(AAZArgAction):  # pylint: disable=abstract-method

    @classmethod
    def setup_operations(cls, dest_ops, values, prefix_keys=None):
        if prefix_keys is None:
            prefix_keys = []
        if values is None or values == []:
            if cls._schema._blank == AAZUndefined:
                raise AAZInvalidValueError("argument cannot be blank")
            assert not isinstance(cls._schema._blank, AAZPromptInput), "Prompt input is not supported in " \
                                                                       "CompoundType args."
            dest_ops.add(copy.deepcopy(cls._schema._blank), *prefix_keys)
        else:
            assert isinstance(values, list)
            for _, key_parts, value in cls.decode_values(values):
                schema = cls._schema
                for k in key_parts:
                    schema = schema[k]  # pylint: disable=unsubscriptable-object
                action = schema._build_cmd_action()
                data = action.format_data(value)
                dest_ops.add(data, *prefix_keys, *key_parts)

    @classmethod
    def decode_values(cls, values):
        for v in values:
            key, key_parts, v = cls._str_parser.split_partial_value(v)
            key_parts, schema = cls.get_schema_by_key_items(key_parts, cls._schema)

            try:
                v = cls._decode_value(schema, v)
            except AAZShowHelp as aaz_help:
                aaz_help.schema = cls._schema
                aaz_help.keys = (*key_parts, *aaz_help.keys)
                raise aaz_help

            yield key, key_parts, v

    @staticmethod
    def get_schema_by_key_items(key_items, schema):
        if not key_items:
            return key_items, schema

        valid_key_items = []
        for item in key_items:
            try:
                schema = schema[item]
                valid_key_items.append(item)
            except AAZUnknownFieldError as err:
                from ._arg import AAZObjectArg, AAZListArg
                if not isinstance(schema, AAZObjectArg):
                    raise err
                is_singular = False
                for field_name, field in schema._fields.items():
                    if not isinstance(field, AAZListArg):
                        continue
                    if field.singular_options and item in field.singular_options:
                        valid_key_items.append(field_name)
                        schema = field.Element
                        valid_key_items.append(_ELEMENT_APPEND_KEY)
                        is_singular = True
                        break
                if not is_singular:
                    raise err
        return tuple(valid_key_items), schema

    @classmethod
    def _decode_value(cls, schema, value):  # pylint: disable=unused-argument
        from ._arg import AAZSimpleTypeArg
        from azure.cli.core.util import get_file_json, shell_safe_json_parse, get_file_yaml

        if len(value) == 0:
            # the express "a=" will return the blank value of schema 'a'
            return AAZBlankArgValue

        if isinstance(schema, AAZSimpleTypeArg):
            # simple type
            v = cls._str_parser(value, is_simple=True)
        else:
            # compound type or any type
            # read from file
            path = os.path.expanduser(value)
            if os.path.exists(path):
                if path.endswith('.yml') or path.endswith('.yaml'):
                    # read from yaml file
                    v = get_file_yaml(path)
                else:
                    # read from json file
                    v = get_file_json(path, preserve_order=True)
            else:
                try:
                    v = cls._str_parser(value)
                except AAZInvalidShorthandSyntaxError as shorthand_ex:
                    try:
                        v = shell_safe_json_parse(value, True)
                    except Exception as ex:
                        logger.debug(ex)  # log parse json failed expression
                        raise shorthand_ex  # raise shorthand syntax exception
        return v


class AAZObjectArgAction(AAZCompoundTypeArgAction):

    @classmethod
    def format_data(cls, data):
        if data == AAZBlankArgValue:
            if cls._schema._blank == AAZUndefined:
                raise AAZInvalidValueError("argument value cannot be blank")
            assert not isinstance(cls._schema._blank, AAZPromptInput), "Prompt input is not supported in Object args."
            data = copy.deepcopy(cls._schema._blank)

        if data is None:
            if cls._schema._nullable:
                return data
            raise AAZInvalidValueError("field is not nullable")

        if isinstance(data, dict):
            result = OrderedDict()
            for key, value in data.items():
                action = cls._schema[key]._build_cmd_action()  # pylint: disable=unsubscriptable-object
                try:
                    result[key] = action.format_data(value)
                except AAZInvalidValueError as ex:
                    raise AAZInvalidValueError(f"Invalid '{key}' : {ex}") from ex
            return result

        raise AAZInvalidValueError(f"dict type value expected, got '{data}'({type(data)})")


class AAZDictArgAction(AAZCompoundTypeArgAction):

    @classmethod
    def format_data(cls, data):
        if data == AAZBlankArgValue:
            if cls._schema._blank == AAZUndefined:
                raise AAZInvalidValueError("argument value cannot be blank")
            assert not isinstance(cls._schema._blank, AAZPromptInput), "Prompt input is not supported in Dict args."
            data = copy.deepcopy(cls._schema._blank)

        if data is None:
            if cls._schema._nullable:
                return data
            raise AAZInvalidValueError("field is not nullable")

        if isinstance(data, dict):
            result = OrderedDict()
            action = cls._schema.Element._build_cmd_action()
            for key, value in data.items():
                try:
                    result[key] = action.format_data(value)
                except AAZInvalidValueError as ex:
                    raise AAZInvalidValueError(f"Invalid '{key}' : {ex}") from ex
            return result

        raise AAZInvalidValueError(f"dict type value expected, got '{data}'({type(data)})")


class AAZListArgAction(AAZCompoundTypeArgAction):

    def __call__(self, parser, namespace, values, option_string=None):
        if not isinstance(getattr(namespace, self.dest), AAZArgActionOperations):
            # overwrite existing namespace value which is not an instance of AAZArgActionOperations.
            # especially the default value of argument.
            setattr(namespace, self.dest, AAZArgActionOperations())
        dest_ops = getattr(namespace, self.dest)
        try:
            if self._schema.singular_options and option_string in self._schema.singular_options:
                # if singular option is used then parsed values by element action
                action = self._schema.Element._build_cmd_action()
                if isinstance(values, list) and len(values) > 1:
                    # append element
                    action.setup_operations(dest_ops, values[:1], prefix_keys=[_ELEMENT_APPEND_KEY])
                    # apply on the last element
                    action.setup_operations(dest_ops, values[1:], prefix_keys=[-1])
                else:
                    # append element
                    action.setup_operations(dest_ops, values, prefix_keys=[_ELEMENT_APPEND_KEY])
            else:
                self.setup_operations(dest_ops, values)
        except (ValueError, KeyError) as ex:
            raise azclierror.InvalidArgumentValueError(f"Failed to parse '{option_string}' argument: {ex}") from ex
        except AAZShowHelp as aaz_help:
            aaz_help.keys = (option_string, *aaz_help.keys)
            self.show_aaz_help(parser, aaz_help)

    @classmethod
    def setup_operations(cls, dest_ops, values, prefix_keys=None):
        if prefix_keys is None:
            prefix_keys = []
        if values is None or values == []:
            if cls._schema._blank == AAZUndefined:
                raise AAZInvalidValueError("argument cannot be blank")
            assert not isinstance(cls._schema._blank, AAZPromptInput), "Prompt input is not supported in List args."
            dest_ops.add(copy.deepcopy(cls._schema._blank), *prefix_keys)
        else:
            assert isinstance(values, list)
            ops = []
            try:
                # Standard Expression
                for key, key_parts, value in cls.decode_values(values):
                    schema = cls._schema
                    for k in key_parts:
                        schema = schema[k]  # pylint: disable=unsubscriptable-object
                    action = schema._build_cmd_action()
                    data = action.format_data(value)
                    ops.append((key_parts, data))
            except AAZShowHelp as aaz_help:
                raise aaz_help
            except Exception as ex:  # pylint: disable=broad-except
                # This part of logic is to support Separate Elements Expression which is widely used in current command,
                # such as:
                #       --args val1 val2 val3.
                # The standard expression of it should be:
                #       --args [val1,val2,val3]

                for value in values:
                    key, _, _ = cls._str_parser.split_partial_value(value)
                    if key is not None:
                        # key should always be None
                        raise ex

                element_action = cls._schema.Element._build_cmd_action()
                if not issubclass(element_action, AAZSimpleTypeArgAction):
                    # Separate Elements Expression only supported for simple type element array
                    raise ex

                # dest_ops
                try:
                    element_ops = AAZArgActionOperations()
                    for value in values:
                        element_action.setup_operations(element_ops, value, prefix_keys=[_ELEMENT_APPEND_KEY])
                except AAZShowHelp as aaz_help:
                    aaz_help.schema = cls._schema
                    aaz_help.keys = (0, *aaz_help.keys)
                    raise aaz_help

                elements = []
                for _, data in element_ops._ops:
                    elements.append(data)
                ops = [([], elements)]

            for key_parts, data in ops:
                dest_ops.add(data, *prefix_keys, *key_parts)

    @classmethod
    def format_data(cls, data):
        if data == AAZBlankArgValue:
            if cls._schema._blank == AAZUndefined:
                raise AAZInvalidValueError("argument value cannot be blank")
            assert not isinstance(cls._schema._blank, AAZPromptInput), "Prompt input is not supported in List args."
            data = copy.deepcopy(cls._schema._blank)

        if data is None:
            if cls._schema._nullable:
                return data
            raise AAZInvalidValueError("field is not nullable")

        if isinstance(data, list):
            result = []
            action = cls._schema.Element._build_cmd_action()
            for idx, value in enumerate(data):
                try:
                    result.append(action.format_data(value))
                except AAZInvalidValueError as ex:
                    raise AAZInvalidValueError(f"Invalid at [{idx}] : {ex}") from ex
            return result

        raise AAZInvalidValueError(f"list type value expected, got '{data}'({type(data)})")


class AAZGenericUpdateAction(Action):  # pylint: disable=too-few-public-methods
    DEST = 'generic_update_args'
    ACTION_NAME = None

    def __call__(self, parser, namespace, values, option_string=None):
        if not getattr(namespace, self.DEST, None):
            setattr(namespace, self.DEST, {"actions": []})
        getattr(namespace, self.DEST)["actions"].append((self.ACTION_NAME or option_string, values))


class AAZGenericUpdateForceStringAction(AAZSimpleTypeArgAction):
    DEST = 'generic_update_args'

    def __call__(self, parser, namespace, values, option_string=None):
        dest_ops = AAZArgActionOperations()
        try:
            self.setup_operations(dest_ops, values)
        except (ValueError, KeyError) as ex:
            raise azclierror.InvalidArgumentValueError(f"Failed to parse '{option_string}' argument: {ex}") from ex
        except AAZShowHelp as aaz_help:
            # show help message
            aaz_help.keys = (option_string, *aaz_help.keys)
            self.show_aaz_help(parser, aaz_help)

        if not getattr(namespace, self.DEST, None):
            setattr(namespace, self.DEST, {"actions": []})
        dest_ops.apply(getattr(namespace, self.DEST), dest='force_string')



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from codecs import open
from setuptools import setup, find_packages

VERSION = "2.88.0"

# If we have source, validate that our version numbers match
# This should prevent uploading releases with mismatched versions.
try:
    with open('azure/cli/core/__init__.py', 'r', encoding='utf-8') as f:
        content = f.read()
except OSError:
    pass
else:
    import re
    import sys

    m = re.search(r'__version__\s*=\s*[\'"](.+?)[\'"]', content)
    if not m:
        print('Could not find __version__ in azure/cli/core/__init__.py')
        sys.exit(1)
    if m.group(1) != VERSION:
        print('Expected __version__ = "{}"; found "{}"'.format(VERSION, m.group(1)))
        sys.exit(1)

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Programming Language :: Python :: 3.13',
    'Programming Language :: Python :: 3.14',
    'License :: OSI Approved :: MIT License',
]

DEPENDENCIES = [
    'argcomplete~=3.5.2',
    'azure-cli-telemetry==1.1.0.*',
    'azure-core~=1.39.0',
    'azure-mgmt-core>=1.2.0,<2',
    'cryptography',
    # On Linux, the distribution (Ubuntu, Debian, etc) and version are logged in telemetry
    'distro; sys_platform == "linux"',
    'humanfriendly~=10.0',
    'jmespath',
    'knack~=0.14.0',
    'microsoft-security-utilities-secret-masker~=1.0.0b4',
    'msal-extensions==1.3.1',
    'msal[broker]==1.36.0; sys_platform == "win32"',
    'msal==1.36.0; sys_platform != "win32"',
    'packaging>=20.9',
    # pkginfo>=1.12.0 reads the spec-defined wheel METADATA / unpacked .dist-info
    # layout produced by modern wheel/setuptools (no metadata.json). Required so
    # WheelExtension.get_metadata works without the legacy wheel==0.30.0 artifact.
    'pkginfo>=1.12.0',
    # psutil can't install on cygwin: https://github.com/Azure/azure-cli/issues/9399
    'psutil>=5.9; sys_platform != "cygwin"',
    'PyJWT>=2.1.0',
    'pyopenssl>=17.1.0',  # https://github.com/pyca/pyopenssl/pull/612
    'py-deviceid',
    'requests[socks]',
]

with open('README.rst', 'r', encoding='utf-8') as f:
    README = f.read()

setup(
    name='azure-cli-core',
    version=VERSION,
    description='Microsoft Azure Command-Line Tools Core Module',
    long_description=README,
    license='MIT',
    author='Microsoft Corporation',
    author_email='azpycli@microsoft.com',
    url='https://github.com/Azure/azure-cli',
    zip_safe=False,
    classifiers=CLASSIFIERS,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests", "azure", "azure.cli"]),
    install_requires=DEPENDENCIES,
    python_requires='>=3.10.0',
    package_data={'azure.cli.core': ['auth/landing_pages/*.html', 'commandIndex.latest.json', 'helpIndex.latest.json']}
)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.testsdk import ScenarioTest, ResourceGroupPreparer, live_only


class IoTCentralTest(ScenarioTest):
    @live_only()
    @ResourceGroupPreparer()  # name_prefix not required, but can be useful
    def test_iot_central_app(self, resource_group, resource_group_location):
        app_name = self.create_random_name(prefix='iotc-cli-test', length=24)
        template_app_name = app_name + '-template'
        template_app_display_name = "My Custom App Display Name"
        managed_identity_app_name = app_name + "-mi"
        rg = resource_group
        location = resource_group_location
        template = 'iotc-pnp-preview@1.0.0'
        updatedName = app_name + 'update'

        # Test 'az iot central app create'
        self.cmd('iot central app create -n {0} -g {1} --subdomain {2} --sku {3}'.format(app_name, rg, app_name, 'ST2'), checks=[
            self.check('resourceGroup', rg),
            self.check('location', location),
            self.check('subdomain', app_name),
            self.check('displayName', app_name),
            self.check('sku.name', 'ST2')]).get_output_in_json()

        # Test 'az iot central app create with template and display name'
        self.cmd('iot central app create -n {0} -g {1} --subdomain {2} --template {3} --display-name \"{4}\" --sku {5}'
                 .format(template_app_name, rg, template_app_name, template, template_app_display_name, 'ST1'), checks=[
                     self.check('resourceGroup', rg),
                     self.check('location', location),
                     self.check('subdomain', template_app_name),
                     self.check('displayName', template_app_display_name),
                     self.check('sku.name', 'ST1'),
                     self.check('template', template)])

        # Test 'az iot central app create with system-assigned managed identity'
        self.cmd('iot central app create -n {0} -g {1} --subdomain {2} --sku {3} --mi-system-assigned'
                .format(managed_identity_app_name, rg, managed_identity_app_name, 'ST2'), checks=[
                    self.check('resourceGroup', rg),
                    self.check('location', location),
                    self.check('subdomain', managed_identity_app_name),
                    self.check('displayName', managed_identity_app_name),
                    self.check('sku.name', 'ST2'),
                    self.check('identity.type', 'SystemAssigned')])

        # Test 'az iot central app update'
        self.cmd('iot central app update -n {0} -g {1} --set displayName={2} subdomain={3} sku.name={4}'
                 .format(template_app_name, rg, updatedName, updatedName, 'ST2'), checks=[
                     self.check('resourceGroup', rg),
                     self.check('location', location),
                     self.check('subdomain', updatedName),
                     self.check('displayName', updatedName),
                     self.check('sku.name', 'ST2')])

        # Test 'az iot central app show'
        self.cmd('iot central app show -n {0} -g {1}'.format(app_name, rg), checks=[
            self.check('resourceGroup', rg),
            self.check('location', location),
            self.check('subdomain', app_name),
            self.check('displayName', app_name),
            self.check('sku.name', 'ST2')])

        # Test 'az iot central app show with template and display name'
        self.cmd('iot central app show -n {0} -g {1}'.format(template_app_name, rg), checks=[
            self.check('resourceGroup', rg),
            self.check('location', location),
            self.check('subdomain', updatedName),
            self.check('displayName', updatedName),
            self.check('sku.name', 'ST2'),
            self.check('template', template)])

        # Test 'az iot central identity show with no managed identity'
        self.cmd('iot central app identity show -n {0} -g {1}'
                .format(app_name, rg), checks=[
                    self.check('type', 'None')])

        # Test 'az iot central identity assign with system-assigned identity'
        self.cmd('iot central app identity assign -n {0} -g {1} --system-assigned'
                .format(managed_identity_app_name, rg), checks=[
                    self.check('type', 'SystemAssigned')])

        # Test 'az iot central identity show with system-assigned identity'
        self.cmd('iot central app identity show -n {0} -g {1}'
                .format(managed_identity_app_name, rg), checks=[
                    self.check('type', 'SystemAssigned')])

        # Test 'az iot central identity remove with system-assigned identity'
        self.cmd('iot central app identity remove -n {0} -g {1} --system-assigned'
                .format(managed_identity_app_name, rg), checks=[
                    self.check('type', 'None')])

        # Test 'az iot central app delete with template and display name'
        self.cmd('iot central app delete -n {0} -g {1} --yes'.format(template_app_name, rg), checks=[
            self.is_empty()])

        # Test 'az iot central app delete with system-assigned identity'
        self.cmd('iot central app delete -n {0} -g {1} --yes'.format(managed_identity_app_name, rg), checks=[
            self.is_empty()])

        # Test 'az iot central app list'
        self.cmd('iot central app list -g {0}'.format(rg), checks=[
            self.check('length([*])', 1),
            self.check('[0].resourceGroup', rg),
            self.check('[0].location', location),
            self.check('[0].subdomain', app_name),
            self.check('[0].displayName', app_name),
            self.check('[0].sku.name', 'ST2')])

        # Test 'az iot central app delete'
        self.cmd('iot central app delete -n {0} -g {1} --yes'.format(app_name, rg), checks=[
            self.is_empty()])

    @live_only()
    @ResourceGroupPreparer()  # name_prefix not required, but can be useful
    def test_iot_central_private_link_and_private_endpoint(self, resource_group):
        name = self.create_random_name(prefix='iotc-cli-test', length=24)
        self.kwargs.update({
            'app_name': name,
            'loc': 'westus',
            'vnet': self.create_random_name('cli-vnet-', 24),
            'subnet': self.create_random_name('cli-subnet-', 24),
            'pe': self.create_random_name('cli-pe-', 24),
            'pe_connection': self.create_random_name('cli-pec-', 24),
            'rg' : resource_group,
            'sku' : 'ST2',
            'type' : 'Microsoft.IoTCentral/iotApps',
            'approve_description' : 'Approved!',
            'reject_description' : 'Rejected!'
        })

        # Setup for Tests

        # Create an iotc app
        result = self.cmd('iot central app create -n {app_name} -g {rg} --subdomain {app_name} --sku {sku}').get_output_in_json()
        self.kwargs['iotc_id'] = result['id']

        # Prepare network for private endpoint connection
        self.cmd('network vnet create -n {vnet} -g {rg} -l {loc} --subnet-name {subnet}',
                        checks=self.check('length(newVNet.subnets)', 1))
        self.cmd('network vnet subnet update -n {subnet} --vnet-name {vnet} -g {rg} '
                '--disable-private-endpoint-network-policies true',
                checks=self.check('privateEndpointNetworkPolicies', 'Disabled'))

        # Test Private Link Resource

        # Test `az iot central app private-link-resource list` with app name and resource group 
        self.cmd('iot central app private-link-resource list -n {app_name} -g {rg} --type {type}', checks=[
            self.check('length(@)', 1)])

        # Test 'az iot central app private-link-resource list` with private link resource id
        self.cmd('iot central app private-link-resource list --id {iotc_id}', checks=[
            self.check('length(@)', 1)])
        

        # Test Private Endpoint Connection

        # Create a private endpoint connection
        pr = self.cmd('az iot central app private-link-resource list -n {app_name} -g {rg} --type {type}').get_output_in_json()
        self.kwargs['group_id'] = pr[0]['groupId']
        self.kwargs['iotc_pr_id'] = pr[0]['id']

        private_endpoint = self.cmd(
            'network private-endpoint create -g {rg} -n {pe} --vnet-name {vnet} --subnet {subnet} -l {loc} '
            '--connection-name {pe_connection} --private-connection-resource-id {iotc_id} '
            '--group-id {group_id}').get_output_in_json()
        self.assertEqual(private_endpoint['name'], self.kwargs['pe'])
        self.assertEqual(private_endpoint['privateLinkServiceConnections'][0]['name'], self.kwargs['pe_connection'])
        self.assertEqual(private_endpoint['privateLinkServiceConnections'][0]['privateLinkServiceConnectionState']['status'], 'Approved')
        self.assertEqual(private_endpoint['privateLinkServiceConnections'][0]['provisioningState'], 'Succeeded')
        self.assertEqual(private_endpoint['privateLinkServiceConnections'][0]['groupIds'][0], self.kwargs['group_id'])
        self.kwargs['pe_id'] = private_endpoint['privateLinkServiceConnections'][0]['id']

        # Show the connection at iot central app
        iotcApp = self.cmd('iot central app show -n {app_name} -g {rg}').get_output_in_json()
        self.assertIn('privateEndpointConnections', iotcApp)
        self.assertEqual(len(iotcApp['privateEndpointConnections']), 1)
        self.assertEqual(iotcApp['privateEndpointConnections'][0]['privateLinkServiceConnectionState']['status'],
                         'Approved')

        self.kwargs['iotc_pec_id'] = iotcApp['privateEndpointConnections'][0]['id']
        
        self.kwargs['iotc_pec_name'] = iotcApp['privateEndpointConnections'][0]['name']

        self.cmd('az iot central app private-link-resource show -n {app_name} -g {rg} --group-id {group_id}',
                 checks=self.check('id','{iotc_pr_id}'))
        
        self.cmd('az iot central app private-link-resource show --id {iotc_pr_id}',
                 checks=self.check('id','{iotc_pr_id}'))

        self.cmd('iot central app private-endpoint-connection show --id {iotc_pec_id}',
                 checks=self.check('id', '{iotc_pec_id}'))

        self.cmd('iot central app private-endpoint-connection list --account-name {app_name} --resource-group {rg}',
                 checks=self.check('length(@)', 1))

        self.cmd('iot central app private-endpoint-connection list --id {iotc_pec_id}',
                 checks=self.check('length(@)', 1))

        self.cmd('iot central app private-endpoint-connection show --account-name {app_name} --name {iotc_pec_name} --resource-group {rg}',
                 checks=self.check('name', '{iotc_pec_name}'))

        self.cmd('iot central app private-endpoint-connection show --account-name {app_name} -n {iotc_pec_name} -g {rg}',
                 checks=self.check('name', '{iotc_pec_name}'))

        self.cmd('iot central app private-endpoint-connection approve --account-name {app_name} -g {rg} --name {iotc_pec_name} --description {approve_description}',
                 checks=[self.check('privateLinkServiceConnectionState.status', 'Approved')])

        # self.cmd('iot central app private-endpoint-connection approve --id {iotc_pec_id}',
        #          checks=[self.check('privateLinkServiceConnectionState.status', 'Approved')])

        self.cmd('iot central app private-endpoint-connection reject --account-name {app_name} -g {rg} --name {iotc_pec_name} --description {reject_description}',
                 checks=[self.check('privateLinkServiceConnectionState.status', 'Rejected')])

        # self.cmd('iot central app private-endpoint-connection reject --id {iotc_pec_id}',
        #          checks=[self.check('privateLinkServiceConnectionState.status', 'Rejected')])

        # with self.assertRaisesRegex(CloudError, 'You cannot approve the connection request after rejection.'):
        #     self.cmd('iot central app private-endpoint-connection approve --account-name {app_name} -g {rg} --name {iotc_pec_name}')

        self.cmd('iot central app private-endpoint-connection delete --id {iotc_pec_id} -y')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import json
from azure.cli.command_modules.acs.azuremonitormetrics.constants import ALERTS_API, RULES_API
from azure.cli.command_modules.acs.azuremonitormetrics.recordingrules.common import truncate_rule_group_name
from knack.util import CLIError


# pylint: disable=line-too-long
def get_recording_rules_template(cmd, azure_monitor_workspace_resource_id):
    from azure.cli.core.util import send_raw_request
    headers = ['User-Agent=azuremonitormetrics.get_recording_rules_template']
    armendpoint = cmd.cli_ctx.cloud.endpoints.resource_manager
    url = f"{armendpoint}{azure_monitor_workspace_resource_id}/providers/microsoft.alertsManagement/alertRuleRecommendations?api-version={ALERTS_API}"
    r = send_raw_request(cmd.cli_ctx, "GET", url, headers=headers)
    data = json.loads(r.text)

    filtered_templates = [
        template for template in data.get('value', [])
        if template.get("properties", {}).get("alertRuleType", "").lower() == "microsoft.alertsmanagement/prometheusrulegroups" and isinstance(template.get("properties", {}).get("rulesArmTemplate", {}).get("resources"), list) and all(
            isinstance(rule, dict) and "record" in rule and "expression" in rule
            for resource in template["properties"]["rulesArmTemplate"]["resources"]
            if resource.get("type", "").lower() == "microsoft.alertsmanagement/prometheusrulegroups"
            for rule in resource.get("properties", {}).get("rules", [])
        )
    ]

    return filtered_templates


# pylint: disable=line-too-long
def put_rules(cmd, default_rule_group_id, default_rule_group_name, mac_region, cluster_resource_id, azure_monitor_workspace_resource_id, cluster_name, default_rules_template, url, enable_rules, i):
    from azure.cli.command_modules.acs._client_factory import get_resources_client
    resources = get_resources_client(cmd.cli_ctx, cmd.cli_ctx.data.get('subscription_id'))
    body = {
        "id": default_rule_group_id,
        "name": default_rule_group_name,
        "type": "Microsoft.AlertsManagement/prometheusRuleGroups",
        "location": mac_region,
        "properties": {
            "scopes": [
                azure_monitor_workspace_resource_id,
                cluster_resource_id
            ],
            "enabled": enable_rules,
            "clusterName": cluster_name,
            "interval": "PT1M",
            "rules": default_rules_template[i]["properties"]["rulesArmTemplate"]["resources"][0]["properties"]["rules"]
        }
    }
    for _ in range(3):
        try:
            resources.begin_create_or_update_by_id(
                default_rule_group_id,
                url.split('api-version=')[1],
                body
            )
            break
        except Exception as e:
            raise CLIError(e)


# pylint: disable=line-too-long
def create_rules(cmd, cluster_subscription, cluster_resource_group_name, cluster_name, azure_monitor_workspace_resource_id, mac_region, raw_parameters):
    default_rules_template = get_recording_rules_template(cmd, azure_monitor_workspace_resource_id)

    cluster_resource_id = (
        f"/subscriptions/{cluster_subscription}/resourceGroups/{cluster_resource_group_name}/providers/Microsoft.ContainerService/managedClusters/{cluster_name}"
    )

    enable_windows_recording_rules = raw_parameters.get("enable_windows_recording_rules", False)

    for index, rule_template in enumerate(default_rules_template):
        rule_name = rule_template["name"]
        is_windows_rule = "win" in rule_name.lower()

        # Determine whether the rule group should be enabled:
        # - If the rule is a Windows rule AND windows recording rules are NOT enabled → disable the rule group (enable_rules = False)
        # - If the rule is a Windows rule AND windows recording rules are enabled → enable the rule group (enable_rules = True)
        # - If the rule is NOT a Windows rule (i.e., a Linux or general rule) → always enable the rule group (enable_rules = True)
        enable_rules = not (is_windows_rule and not enable_windows_recording_rules)

        rule_group_name = truncate_rule_group_name(f"{rule_template['name']}-{cluster_name}")
        rule_group_id = f"/subscriptions/{cluster_subscription}/resourceGroups/{cluster_resource_group_name}/providers/Microsoft.AlertsManagement/prometheusRuleGroups/{rule_group_name}"
        url = f"{cmd.cli_ctx.cloud.endpoints.resource_manager}{rule_group_id}?api-version={RULES_API}"

        put_rules(
            cmd,
            rule_group_id,
            rule_group_name,
            mac_region,
            cluster_resource_id,
            azure_monitor_workspace_resource_id,
            cluster_name,
            default_rules_template,
            url,
            enable_rules,
            index
        )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long
import unittest
import os
import tempfile

from azure.cli.core.commands.validators import validate_file_or_dict


class TestValidators(unittest.TestCase):

    def test_validate_file_or_dict(self):
        # verify user folder is expanded before load the file
        temp_name = next(tempfile._get_candidate_names())
        file_path = '~/' + temp_name
        local_file_path = os.path.expanduser(file_path)
        with open(local_file_path, 'w') as f:
            f.write('{"prop":"val"}')

        # verify we load the json content correctly
        try:
            res = validate_file_or_dict(file_path)
            self.assertEqual(res['prop'], "val")
        finally:
            os.remove(local_file_path)

        # verify expanduser call won't mess up the json data
        data = '{"~d": "~/haha"}'
        res = validate_file_or_dict(data)
        self.assertEqual(res['~d'], '~/haha')

        # verify expanduser call again, but use single quot
        data = "{'~d': '~/haha'}"
        res = validate_file_or_dict(data)
        self.assertEqual(res['~d'], '~/haha')


if __name__ == '__main__':
    unittest.main()
